! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

! In this module:
!  subroutine read_all_gti_eventfile(EventFile, status, reopen)
!  subroutine free_all_gti_eventfile(EventFile, status)
!  subroutine free_all_qualgti_eventfile(EventFile, status)
!  subroutine free_all_flaregti_eventfile(EventFile, status)
!  subroutine read_all_qualgti_eventfile(EventFile, status, reopen)
!  subroutine read_all_flaregti_eventfile(EventFile, status, reopen)
!  subroutine get_merged_events_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_gti_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_qualgti_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_flaregti_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merge_std_header_eventfile(EventFile, EventFile_out, status)
!  subroutine merge_stdheader_eventfile(EventFile_in, EventFile_out, status)
!  subroutine compare_stdheader_eventfile(EventFile, status, alwayswarn)
!  subroutine get_merged_extension_eventfile(EventFile, EventFile_out, status)
!  subroutine get_merged_extension_eventfile_reopen(EventFile, EventFile_out, status)
!  subroutine read_all_hk_eventfile(EventFile, status, reopen)
!  subroutine read_all_hkx1_eventfile(EventFile, status, reopen)
!  subroutine read_all_hkx2_eventfile(EventFile, status, reopen)
!  subroutine read_all_hkx3_eventfile(EventFile, status, reopen)
!  subroutine read_all_hkx5_eventfile(EventFile, status, reopen)
!  subroutine get_merged_hk_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_hkx1_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_hkx5_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_hkx2_eventfile(EventFile_in, EventFile_out, status)
!  subroutine get_merged_hkx3_eventfile(EventFile_in, EventFile_out, status)
!  subroutine read_all_offset_eventfile(EventFile, status, reopen)
!  subroutine get_merged_offset_eventfile(EventFile_in, EventFile_out, status)
!  subroutine read_all_badpix_eventfile(EventFile, status, reopen)
!  subroutine get_merged_badpix_eventfile(EventFile_in, EventFile_out, status)
!  subroutine read_all_deadcor_eventfile(EventFile, status, reopen)
!  subroutine read_all_events_eventfile(EventFile, status, reopen)
!  subroutine get_merged_deadcor_eventfile(EventFile_in, EventFile_out, status)
!  subroutine read_all_corrat_eventfile(EventFile, status, reopen)
!  subroutine get_merged_corrat_eventfile(EventFile_in, EventFile_out, status)
!  subroutine free_all_corrat_eventfile(EventFile, status)
!  subroutine free_all_deadcor_eventfile(EventFile, status)
!  subroutine free_all_events_eventfile(EventFile, status)
!  subroutine free_all_badpix_eventfile(EventFile, status)
!  subroutine free_all_hk_eventfile(EventFile, status)
!  subroutine free_all_hkx1_eventfile(EventFile, status)
!  subroutine free_all_hkx2_eventfile(EventFile, status)
!  subroutine free_all_hkx3_eventfile(EventFile, status)
!  subroutine free_all_hkx5_eventfile(EventFile, status)
!  subroutine free_all_offset_eventfile(EventFile, status)
!  subroutine merge_ccdnr(array1, array2, outarray, array1_str, array2_str, outarray_str, status)
!  subroutine compare_rawxy_masks(ccdnrs1, rawxy_mask1, ccdnrs2, rawxy_mask2, status)
!  subroutine merge_rawxy_masks(ccdnrs2, rawxy_mask2, rawxy_mask, ccd_is_merged, status)

module evtool_eventfile_routines_mod

  use erolib_general_mod
  use evtool_deadcor_mod
  use evtool_corrat_mod
  use erolib_region_mod
  use evtool_gti_mod
  use evtool_eventfile_mod
  use time_mod

  implicit none

logical, private, parameter :: doTest=.false.

contains

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the gti extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_gti_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_gti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_gti(status, reopen)
      else
        call EventFile(ii)%read_gti(status)
      end if
    end do

  end subroutine read_all_gti_eventfile


      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free all the gti extensions of an array of eventfiles
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_gti_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_gti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_gti(status)
    end do

  end subroutine free_all_gti_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free all the qualgti extensions of an array of eventfiles
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_qualgti_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_qualgti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_qualgti(status)
    end do

  end subroutine free_all_qualgti_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free all the flaregti extensions of an array of eventfiles
  !
  !
  !
  !> @param[in] EventFile
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_flaregti_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_flaregti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_flaregti(status)
    end do

  end subroutine free_all_flaregti_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the qualgti extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_qualgti_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_qualgti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_qualgti(status, reopen)
      else
        call EventFile(ii)%read_qualgti(status)
      end if
    end do

  end subroutine read_all_qualgti_eventfile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the flaregti extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_flaregti_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_flaregti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_flaregti(status, reopen, convert=true)
      else
        call EventFile(ii)%read_flaregti(status, convert=true)
      end if
    end do

  end subroutine read_all_flaregti_eventfile


    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge the events extension of an array of tEventFile structures into one
  !> all initial filters and "keys" will be taken from the first input EventFile
  !> except for the ccdnr filter which will be merged
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_events_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, nfiles

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tEvents), dimension(size(EventFile_in)) :: tmp

    character(*), parameter :: function = "get_merged_events_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the events extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_events == false) then
        call EventFile_out%events%init(0, status)
        EventFile_out%ext_events = false
        return
      end if
    end do
    EventFile_out%ext_events = true

    if(nfiles == 1) then
      EventFile_out%events = EventFile_in(1)%events
      return
    end if

    do ii = 1, nfiles
      tmp(ii) = EventFile_in(ii)%events
    end do

    call merge_events(tmp, EventFile_out%events, status)
    call EventFile_out%delete_duplicates_events(status)

  end subroutine get_merged_events_eventfile



  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the merged GTI for all GTIs in the infiles relevant for the outfile
  !
  !> @param[in] EventFile
  !> @param[out] file
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_gti_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, first, nfiles

    type(tEventFile), intent(inout) :: EventFile_out
    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tGTI) :: gti_tmp

    character(*), parameter :: function = "get_merged_gti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the gti extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_gti == false) then
        call alloc_gti(EventFile_out%gti, 0, status)
        EventFile_out%ext_gti = false
        return
      end if
    end do
    EventFile_out%ext_gti = true

    call alloc_gti(EventFile_out%gti, EventFile_out%ninst, status)

    ! loop over all possible gti extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%gti(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, size(EventFile_in)
        if(size(EventFile_in(jj)%gti) == 0) cycle
        ! loop over their corresponding gti extensions
        do kk = 1, size(EventFile_in(jj)%gti)
          ! check if the infile gti extension fits the out corrat extension
          if(EventFile_in(jj)%gti(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%gti(ii) = EventFile_in(jj)%gti(kk)
            else
              gti_tmp = EventFile_out%gti(ii)
              call merge_gti(gti_tmp, EventFile_in(jj)%gti(kk), EventFile_out%gti(ii), or, status)
            end if
          end if
        end do
      end do
    end do
    call gti_tmp%free(status)

  end subroutine get_merged_gti_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the merged QUALGTI for all QUALGTI in the infiles relevant for the outfile
  !
  !> @param[in] EventFile
  !> @param[out] file
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_qualgti_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, first, nfiles

    type(tEventFile), intent(inout) :: EventFile_out
    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tGTI) :: qualgti_tmp

    character(*), parameter :: function = "get_merged_qualgti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the qualgti extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_qualgti == false) then
        call alloc_gti(EventFile_out%qualgti, 0, status)
        EventFile_out%ext_qualgti = false
        return
      end if
    end do
    EventFile_out%ext_qualgti = true

    call alloc_gti(EventFile_out%qualgti, EventFile_out%ninst, status)

    ! loop over all possible corrat extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%qualgti(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, size(EventFile_in)
        !                if(size(EventFile_in(jj)%qualgti) == 0) cycle
        ! loop over their corresponding qualgti extensions
        do kk = 1, size(EventFile_in(jj)%qualgti)
          ! check if the infile qualgti extension fits the out corrat extension
          if(EventFile_in(jj)%qualgti(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%qualgti(ii) = EventFile_in(jj)%qualgti(kk)
            else
              qualgti_tmp = EventFile_out%qualgti(ii)
              call merge_gti(qualgti_tmp, EventFile_in(jj)%qualgti(kk), EventFile_out%qualgti(ii), or, status)
            end if
          end if
        end do
      end do
    end do

    call qualgti_tmp%free(status)

  end subroutine get_merged_qualgti_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> get the merged GTI for all FLAREGTIs in the infiles relevant for the outfile
  !
  !> @param[in] EventFile_out
  !> @param[out] EventFile_in
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_flaregti_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, first, nfiles

    type(tEventFile), intent(inout) :: EventFile_out
    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tGTI) :: flaregti_tmp

    character(*), parameter :: function = "get_merged_flaregti_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the qualgti extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_flaregti == false) then
        call alloc_gti(EventFile_out%flaregti, 0, status)
        EventFile_out%ext_flaregti = false
        return
      end if
    end do
    EventFile_out%ext_flaregti = true

    call alloc_gti(EventFile_out%flaregti, EventFile_out%ninst, status)

    ! loop over all possible corrat extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%flaregti(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, size(EventFile_in)
        ! loop over their corresponding qualgti extensions
        do kk = 1, size(EventFile_in(jj)%flaregti)
          ! check if the infile qualgti extension fits the out corrat extension
          if(EventFile_in(jj)%flaregti(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%flaregti(ii) = EventFile_in(jj)%flaregti(kk)
            else
              flaregti_tmp = EventFile_out%flaregti(ii)
              call merge_gti(flaregti_tmp, EventFile_in(jj)%flaregti(kk), EventFile_out%flaregti(ii), or, status)
            end if
          end if
        end do
      end do
    end do

    call flaregti_tmp%free(status)

  end subroutine get_merged_flaregti_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> initialize and merge all std_header keywords of multiple EventFiles and insert them into a single EventFile
  !
  !
  !
  !> @param[in] EventFile
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

!*** 2019-09-11 IMS: does not seem to be used anywhere.

  subroutine get_merge_std_header_eventfile(EventFile, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    type(tEventFile), dimension(:), intent(inout) :: EventFile
    type(tEventFile), intent(inout) :: EventFile_out

    character(*), parameter :: function = "get_merge_std_header_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call read_events_ext_keywords_eventfile(EventFile, status)
    call compare_stdheader_eventfile(EventFile, status)
    call merge_stdheader_eventfile(EventFile, EventFile_out, status)

  end subroutine get_merge_std_header_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge the stdheader of Eventfiles into one
  !
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_stdheader_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status

    character(*), parameter :: function = "merge_stdheader_eventfile"

    type(tEventFile), dimension(:), intent(in) :: EventFile_in
    type(tEventFile), intent(out) :: EventFile_out

    integer :: ii, same

    ! allocatable
    character(buffsize), dimension(:), allocatable :: instrum
    real(r8), dimension(:), allocatable :: emin, emax
    integer, dimension(:), allocatable :: ccdnr_infile
    logical, dimension(ntel) :: ccd_is_merged

    if(status > 100 .or. status == s_errRep) return

    ! make sanity check
    if(size(EventFile_in) == 0) then
      call wrerr(function, s_err, "number of input eventfiles for merging = 0")
      status = s_errRep
  return
    end if

    ! do not completely copy the EventFile but just the important parameters
    !        EventFile_out = EventFile_in(1)
    EventFile_out%ninst = EventFile_in(1)%ninst
    EventFile_out%ccdnr_infile = EventFile_in(1)%ccdnr_infile
    EventFile_out%instrum = EventFile_in(1)%instrum

    EventFile_out%filter_pattern = EventFile_in(1)%filter_pattern
    EventFile_out%filter_energy = EventFile_in(1)%filter_energy
    EventFile_out%filter_telid = EventFile_in(1)%filter_telid
    EventFile_out%filter_gti = EventFile_in(1)%filter_gti
    EventFile_out%filter_flag = EventFile_in(1)%filter_flag
    EventFile_out%filter_region = EventFile_in(1)%filter_region
    EventFile_out%filter_rawxy = EventFile_in(1)%filter_rawxy

    EventFile_out%rawxy_mask = EventFile_in(1)%rawxy_mask

    EventFile_out%std_header = EventFile_in(1)%std_header

    if(size(EventFile_in) == 1) return

    do ii = 2, size(EventFile_in)
      call merge_ccdnr(EventFile_out%ccdnr_infile, EventFile_in(ii)%ccdnr_infile, ccdnr_infile, &
        EventFile_out%instrum, EventFile_in(ii)%instrum, instrum, status)

      EventFile_out%ccdnr_infile = ccdnr_infile
      EventFile_out%instrum = instrum
      ! update the std header
      EventFile_out%std_header%ninst = size(ccdnr_infile)
      EventFile_out%std_header%instrumx(:size(ccdnr_infile)) = instrum(:size(ccdnr_infile))
      if(size(ccdnr_infile) > 1) then
        EventFile_out%std_header%instrume = "merged"
      end if

      call compare_str(EventFile_out%std_header%filter, EventFile_in(ii)%std_header%filter, same, status, true)
      if(same == false) then
        EventFile_out%std_header%filter = "mixed"
        call wrerr(function, warn, "filter keyword will be set to mixed")
      end if

      EventFile_out%std_header%pat_sel = ior(EventFile_out%std_header%pat_sel, EventFile_in(ii)%std_header%pat_sel)
      EventFile_out%std_header%flagsel = iand(EventFile_out%std_header%flagsel, EventFile_in(ii)%std_header%flagsel)

      call merge_or_r8(EventFile_out%std_header%eminx, EventFile_out%std_header%emaxx, EventFile_out%std_header%nenergy, &
        EventFile_in(ii)%std_header%eminx, EventFile_in(ii)%std_header%emaxx, EventFile_in(ii)%std_header%nenergy, &
        emin, emax, status)

      EventFile_out%std_header%nenergy = size(emin)

      if(size(emin) > size(EventFile_out%std_header%eminx)) then
        call wrerr(function, warn, "merged file would contain more energy intervals, number of energy intervals are truncated")
      end if

      EventFile_out%std_header%eminx(:size(emin)) = emin
      EventFile_out%std_header%emaxx(:size(emax)) = emax

      ! update the tstart and tstop
      if(EventFile_in(ii)%std_header%tstart < EventFile_out%std_header%tstart) EventFile_out%std_header%tstart = EventFile_in(ii)%std_header%tstart
      if(EventFile_in(ii)%std_header%tstop > EventFile_out%std_header%tstop) EventFile_out%std_header%tstop = EventFile_in(ii)%std_header%tstop
      call sec2date(EventFile_out%std_header%tstart, EventFile_out%std_header%mjdref, EventFile_out%std_header%date_obs, status)
      call sec2date(EventFile_out%std_header%tstop, EventFile_out%std_header%mjdref, EventFile_out%std_header%date_end, status)

      ! GTI
      if(EventFile_in(ii)%std_header%gti_sel /= EventFile_out%std_header%gti_sel) then
        EventFile_out%std_header%gti_sel = "undef"
      end if

      ! compare skyfields
      if(trim(EventFile_out%std_header%skyfield) .ne. trim(EventFile_in(ii)%std_header%skyfield)) then
        EventFile_out%std_header%skyfield = "merged"
      end if

    end do

    do ii=1,ntel
      ccd_is_merged(ii) = .false. ! the default value
    end do

    ! Merge the RAWXY masks last, because we need the full list of CCD numbers. Note that we do this here rather than in get_merged_extension_eventfile() because we need the full RAWXY extension before we apply the current RAWXY filter to it in update_filter_settings_eventfile().
    !
    do ii = 2, size(EventFile_in)
      call merge_rawxy_masks(EventFile_in(ii)%ccdnr_infile, EventFile_in(ii)%rawxy_mask, &
        EventFile_out%rawxy_mask, ccd_is_merged, status)
    end do
    if(allocated(EventFile_out%rawxy_mask%values))&
      EventFile_out%filter_rawxy = .true.

    ! set the ninst for the whole outfile
    EventFile_out%ninst = EventFile_out%std_header%ninst

    call dealloc(ccdnr_infile, status)
    call dealloc(instrum, status)
    call dealloc(emin, status)
    call dealloc(emax, status)

  end subroutine merge_stdheader_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> compare the stdheader keywords and if option is set produces warnings for each inconsistency
  !
  !
  !
  !> @param[in] EventFile
  !> @param[out]
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine compare_stdheader_eventfile(EventFile, status, alwayswarn)
    implicit none

    integer :: ii, warnlevel, same
    integer, intent(inout) :: status
    integer, intent(in), optional :: alwayswarn

    integer :: compare_emin, compare_emax, nenergy, nenergy_1

    type(tEventFile), dimension(:), intent(in) :: EventFile

    character(*), parameter :: function = "compare_stdheader_eventfile"
    character(buffsize) :: warnmsg

    if(status > 100 .or. status == s_errRep) return

    if(size(EventFile) < 1) then
      call wrerr(function, s_err, "EventFile array is empty")
      status = s_errRep
      return
    end if

    if(size(EventFile) == 1) return

    warnlevel = 0
    if(present(alwayswarn)) then
      if(alwayswarn == true) warnlevel = ibset(warnlevel, 8)
    end if

    do ii = 2, size(EventFile)
      if(compare(EventFile(ii)%ccdnr_infile, EventFile(1)%ccdnr_infile) /= true) then
        warnlevel = ibset(warnlevel, 0)
        warnmsg = "instruments in input files are different"
        if(ibclr(warnlevel, 0) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
          return
        end if
      end if

      if(EventFile(ii)%std_header%pat_sel /= EventFile(1)%std_header%pat_sel) then
        warnlevel = ibset(warnlevel, 1)
        warnmsg = "pat_sel in input files are different"
        if(ibclr(warnlevel, 1) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
          return
        end if
      end if

      if(EventFile(ii)%std_header%flagsel /= EventFile(1)%std_header%flagsel) then
        warnlevel = ibset(warnlevel, 2)
        warnmsg = "flagsel in input files are different"
        if(ibclr(warnlevel, 2) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
          return
        end if
      end if

      if(EventFile(ii)%std_header%gti_sel /= EventFile(1)%std_header%gti_sel) then
        warnmsg = "gti_sel in input files are different"
        call wrerr(function, status, warnmsg)
!        status = s_errRep
      end if

      call compare_rawxy_masks(EventFile(1)%ccdnr_infile, EventFile(1)%rawxy_mask,&
        EventFile(ii)%ccdnr_infile, EventFile(ii)%rawxy_mask, status)

      nenergy = EventFile(ii)%std_header%nenergy
      nenergy_1 = EventFile(1)%std_header%nenergy
      compare_emin = compare(EventFile(ii)%std_header%eminx(:nenergy), EventFile(1)%std_header%eminx(:nenergy_1))
      compare_emax = compare(EventFile(ii)%std_header%emaxx(:nenergy), EventFile(1)%std_header%emaxx(:nenergy_1))

      if(compare_emin == false .or. compare_emax == false) then
        warnlevel = ibset(warnlevel, 3)
        warnmsg = "emin / emax in input files are different"
        if(ibclr(warnlevel, 3) > 0) then
          call wrerr(function, status, warnmsg)
          status = s_errRep
          return
        end if
      end if

        ! compare std_header
      call compare_str(EventFile(1)%std_header%obs_mode, EventFile(ii)%std_header%obs_mode, same, status, true)
      if(same == false) call wrerr(function, warn, "Different OBS_MODE in infiles")
      call compare_str(EventFile(1)%std_header%datamode, EventFile(ii)%std_header%datamode, same, status, true)
      if(same == false) call wrerr(function, warn, "Different DATAMODE in infiles")
      call compare_str(EventFile(1)%std_header%filter, EventFile(ii)%std_header%filter, same, status, true)
      if(same == false) call wrerr(function, warn, "Different FILTER in infiles")

      !      if(EventFile(1)%std_header%obs_mode /= EventFile(ii)%std_header%obs_mode) call wrerr(function, warn, "Different OBS_MODE in infiles")
      !      if(EventFile(1)%std_header%datamode /= EventFile(ii)%std_header%datamode) call wrerr(function, warn, "Different DATAMODE in infiles")
      ! TODO implement filter comparison if available
      !            if(EventFile(1)%std_header%filter /= EventFile(ii)%std_header%filter) call wrerr(function, warn, "Different FILTER in infiles")


      if(abs(EventFile(1)%std_header%frametim - EventFile(ii)%std_header%frametim) > tiny(EventFile(1)%std_header%frametim)) then
        call wrerr(function, warn, "Different FRAMETIM in infiles")
      end if

      if(EventFile(1)%std_header%radecsys /= EventFile(ii)%std_header%radecsys) then
        call wrerr(function, s_err, "Different RADECSYS in infiles")
        status = s_errRep
        return
      end if

      if(abs(EventFile(1)%std_header%equinox - EventFile(ii)%std_header%equinox) > tiny(EventFile(1)%std_header%equinox)) then
        call wrerr(function, s_err, "Different EQUINOX in infiles")
        status = s_errRep
        return
      end if

      if(abs(EventFile(1)%std_header%timezero - EventFile(ii)%std_header%timezero) > tiny(EventFile(1)%std_header%timezero)) then
        call wrerr(function, s_err, "Different TIMEZERO in infiles")
        status = s_errRep
        return
      end if

      if(EventFile(1)%std_header%timeunit /= EventFile(ii)%std_header%timeunit) then
        call wrerr(function, s_err, "Different TIMEUNIT in infiles")
        status = s_errRep
        return
      end if

      if(EventFile(1)%std_header%timesys /= EventFile(ii)%std_header%timesys) then
        call wrerr(function, s_err, "Different TIMESYS in infiles")
        status = s_errRep
        return
      end if

    end do

  end subroutine compare_stdheader_eventfile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> initialize and merge all extensions of the EventFiles and insert them into a single EventFile
  !
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_extension_eventfile(EventFile, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status

    type(tEventFile), dimension(:), intent(inout) :: EventFile
    type(tEventFile), intent(inout) :: EventFile_out

    character(*), parameter :: function = "get_merged_extension_eventfile"

    if(status > 100 .or. status == s_errRep) return

    call get_merged_events_eventfile(EventFile, EventFile_out, status)
    call free_all_events_eventfile(EventFile, status)

    call get_merged_badpix_eventfile(EventFile, EventFile_out, status)
    call free_all_badpix_eventfile(EventFile, status)

    call get_merged_corrat_eventfile(EventFile, EventFile_out, status)
    call free_all_corrat_eventfile(EventFile, status)

    call get_merged_deadcor_eventfile(EventFile, EventFile_out, status)
    call free_all_deadcor_eventfile(EventFile, status)

    call get_merged_hk_eventfile(EventFile, EventFile_out, status)
    call free_all_hk_eventfile(EventFile, status)

    call get_merged_hkx1_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx1_eventfile(EventFile, status)

    call get_merged_hkx2_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx2_eventfile(EventFile, status)

    call get_merged_hkx3_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx3_eventfile(EventFile, status)

    call get_merged_hkx5_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx5_eventfile(EventFile, status)

    call get_merged_offset_eventfile(EventFile, EventFile_out, status)
    call free_all_offset_eventfile(EventFile, status)

    call get_merged_gti_eventfile(EventFile, EventFile_out, status)
    call free_all_gti_eventfile(EventFile, status)

    call get_merged_qualgti_eventfile(EventFile, EventFile_out, status)
    call free_all_qualgti_eventfile(EventFile, status)

    call get_merged_flaregti_eventfile(EventFile, EventFile_out, status)
    call free_all_flaregti_eventfile(EventFile, status)

  end subroutine get_merged_extension_eventfile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> initialize and merge all extensions of the EventFiles and insert them into a single EventFile
  !
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_extension_eventfile_reopen(EventFile, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status

    type(tEventFile), dimension(:), intent(inout) :: EventFile
    type(tEventFile), intent(inout) :: EventFile_out

    character(*), parameter :: function = "get_merged_extension_eventfile_reopen"

    if(status > 100 .or. status == s_errRep) return

    call read_all_events_eventfile(EventFile, status, true)
    call get_merged_events_eventfile(EventFile, EventFile_out, status)
    call free_all_events_eventfile(EventFile, status)

    call read_all_badpix_eventfile(EventFile, status, true)
    call get_merged_badpix_eventfile(EventFile, EventFile_out, status)
    call free_all_badpix_eventfile(EventFile, status)

    call read_all_corrat_eventfile(EventFile, status, true)
    call get_merged_corrat_eventfile(EventFile, EventFile_out, status)
    call free_all_corrat_eventfile(EventFile, status)

    call read_all_deadcor_eventfile(EventFile, status, true)
    call get_merged_deadcor_eventfile(EventFile, EventFile_out, status)
    call free_all_deadcor_eventfile(EventFile, status)

    call read_all_hk_eventfile(EventFile, status, true)
    call get_merged_hk_eventfile(EventFile, EventFile_out, status)
    call free_all_hk_eventfile(EventFile, status)

    call read_all_hkx1_eventfile(EventFile, status, true)
    call get_merged_hkx1_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx1_eventfile(EventFile, status)

    call read_all_hkx2_eventfile(EventFile, status, true)
    call get_merged_hkx2_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx2_eventfile(EventFile, status)

    call read_all_hkx3_eventfile(EventFile, status, true)
    call get_merged_hkx3_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx3_eventfile(EventFile, status)

    call read_all_hkx5_eventfile(EventFile, status, true)
    call get_merged_hkx5_eventfile(EventFile, EventFile_out, status)
    call free_all_hkx5_eventfile(EventFile, status)

    call read_all_offset_eventfile(EventFile, status, true)
    call get_merged_offset_eventfile(EventFile, EventFile_out, status)
    call free_all_offset_eventfile(EventFile, status)

    call read_all_gti_eventfile(EventFile, status, true)
    call get_merged_gti_eventfile(EventFile, EventFile_out, status)
    call free_all_gti_eventfile(EventFile, status)

    call read_all_qualgti_eventfile(EventFile, status, true)
    call get_merged_qualgti_eventfile(EventFile, EventFile_out, status)
    call free_all_qualgti_eventfile(EventFile, status)

    call read_all_flaregti_eventfile(EventFile, status, true)
    call get_merged_flaregti_eventfile(EventFile, EventFile_out, status)
    call free_all_flaregti_eventfile(EventFile, status)

    ! rawxy_mask read/merge is done elsewhere.

  end subroutine get_merged_extension_eventfile_reopen

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hk extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_hk_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_hk_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_hk(status, reopen)
      else
        call EventFile(ii)%read_hk(status)
      end if
    end do

  end subroutine read_all_hk_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx1 extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_hkx1_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_hkx1_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_hkx1(status, reopen)
      else
        call EventFile(ii)%read_hkx1(status)
      end if
    end do

  end subroutine read_all_hkx1_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx2 extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_hkx2_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_hkx2_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_hkx2(status, reopen)
      else
        call EventFile(ii)%read_hkx2(status)
      end if
    end do

  end subroutine read_all_hkx2_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx3 extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_hkx3_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_hkx3_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_hkx3(status, reopen)
      else
        call EventFile(ii)%read_hkx3(status)
      end if
    end do

  end subroutine read_all_hkx3_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the hkx5 extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_hkx5_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_hkx5_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_hkx5(status, reopen)
      else
        call EventFile(ii)%read_hkx5(status)
      end if
    end do

  end subroutine read_all_hkx5_eventfile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge hk from infiles to outfiles
  !> also creates hdu of hk in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_hk_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tHK) :: tmp

    character(*), parameter :: function = "get_merged_hk_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    if(status == s_errRep) return

    ! check that each input eventfile has the hk extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_hk == false) then
        call alloc_hk(EventFile_out%hk, 0, status)
        EventFile_out%ext_hk = false
        return
      end if
    end do
    EventFile_out%ext_hk = true


    ! allocate the number of hk extensions in outfile
    call alloc_hk(EventFile_out%hk, EventFile_out%ninst, status)

    ! loop over all possible hk extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%hk(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%hk) == 0) cycle
        ! loop over their corresponding hk extensions
        do kk = 1, size(EventFile_in(jj)%hk)
          ! check if the infile hk extension fits the out hk extension
          if(EventFile_in(jj)%hk(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%hk(ii) = EventFile_in(jj)%hk(kk)
            else
              tmp = EventFile_out%hk(ii)
              call merge_hk([tmp, EventFile_in(jj)%hk(kk)], EventFile_out%hk(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)
    call EventFile_out%delete_duplicates_hk(status)

  end subroutine get_merged_hk_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge hk from infiles to outfiles
  !> also creates hdu of hkx1 in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_hkx1_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tHKx1) :: tmp

    character(*), parameter :: function = "get_merged_hkx1_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    if(status == s_errRep) return !20200107 IMS: wtf???

    ! check that each input eventfile has the hk extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_hkx1 == false) then
        call alloc_hkx1(EventFile_out%hkx1, 0, status)
        EventFile_out%ext_hkx1 = false
  return
      end if
    end do
    EventFile_out%ext_hkx1 = true

    ! allocate the number of hk extensions in outfile
    call alloc_hkx1(EventFile_out%hkx1, EventFile_out%ninst, status)
    ! loop over all possible hk extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%hkx1(ii)%init(0, status) ! sets %extname to the default.
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%hk) == 0) cycle
        ! loop over their corresponding hk extensions
        do kk = 1, size(EventFile_in(jj)%hkx1)
          ! check if the infile hk extension fits the out hk extension
          if(EventFile_in(jj)%hkx1(kk)%tel_id == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%hkx1(ii) = EventFile_in(jj)%hkx1(kk)
            else
              tmp = EventFile_out%hkx1(ii)
              call merge_hkx1(tmp, EventFile_in(jj)%hkx1(kk), EventFile_out%hkx1(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)

      ! TODO add delete duplicates
  !        call EventFile_out%delete_duplicates_hk(status)

  end subroutine get_merged_hkx1_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge hk from infiles to outfiles
  !> also creates hdu of hkx5 in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_hkx5_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tHKx5) :: tmp

    character(*), parameter :: function = "get_merged_hkx5_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    if(status == s_errRep) return

    ! check that each input eventfile has the hk extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_hkx5 == false) then
        call alloc_hkx5(EventFile_out%hkx5, 0, status)
        EventFile_out%ext_hkx5 = false
        return
      end if
    end do
    EventFile_out%ext_hkx5 = true

    ! allocate the number of hk extensions in outfile
    call alloc_hkx5(EventFile_out%hkx5, EventFile_out%ninst, status)

    ! loop over all possible hk extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%hkx5(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%hk) == 0) cycle
        ! loop over their corresponding hk extensions
        do kk = 1, size(EventFile_in(jj)%hkx5)
          ! check if the infile hk extension fits the out hk extension
          if(EventFile_in(jj)%hkx5(kk)%tel_id == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%hkx5(ii) = EventFile_in(jj)%hkx5(kk)
            else
              tmp = EventFile_out%hkx5(ii)
              call merge_hkx5(tmp, EventFile_in(jj)%hkx5(kk), EventFile_out%hkx5(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)

      ! TODO add delete duplicates
  !        call EventFile_out%delete_duplicates_hk(status)

  end subroutine get_merged_hkx5_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge hk from infiles to outfiles
  !> also creates hdu of hkx2 in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_hkx2_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tHKx2) :: tmp

    character(*), parameter :: function = "get_merged_hkx2_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    if(status == s_errRep) return

    ! check that each input eventfile has the hk extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_hkx2 == false) then
        call alloc_hkx2(EventFile_out%hkx2, 0, status)
        EventFile_out%ext_hkx2 = false
        return
      end if
    end do
    EventFile_out%ext_hkx2 = true

    ! allocate the number of hk extensions in outfile
    call alloc_hkx2(EventFile_out%hkx2, EventFile_out%ninst, status)

    ! loop over all possible hk extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%hkx2(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%hk) == 0) cycle
        ! loop over their corresponding hk extensions
        do kk = 1, size(EventFile_in(jj)%hkx2)
          ! check if the infile hk extension fits the out hk extension
          if(EventFile_in(jj)%hkx2(kk)%tel_id == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%hkx2(ii) = EventFile_in(jj)%hkx2(kk)
            else
              tmp = EventFile_out%hkx2(ii)
              call merge_hkx2(tmp, EventFile_in(jj)%hkx2(kk), EventFile_out%hkx2(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)

      ! TODO add delete duplicates
  !        call EventFile_out%delete_duplicates_hk(status)

  end subroutine get_merged_hkx2_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge hk from infiles to outfiles
  !> also creates hdu of hkx3 in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_hkx3_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tHKx3) :: tmp

    character(*), parameter :: function = "get_merged_hkx3_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    if(status == s_errRep) return

    ! check that each input eventfile has the hk extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_hkx3 == false) then
        call alloc_hkx3(EventFile_out%hkx3, 0, status)
        EventFile_out%ext_hkx3 = false
        return
      end if
    end do
    EventFile_out%ext_hkx3 = true

    ! allocate the number of hk extensions in outfile
    call alloc_hkx3(EventFile_out%hkx3, EventFile_out%ninst, status)

    ! loop over all possible hk extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%hkx3(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%hk) == 0) cycle
        ! loop over their corresponding hk extensions
        do kk = 1, size(EventFile_in(jj)%hkx3)
          ! check if the infile hk extension fits the out hk extension
          if(EventFile_in(jj)%hkx3(kk)%tel_id == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%hkx3(ii) = EventFile_in(jj)%hkx3(kk)
            else
              tmp = EventFile_out%hkx3(ii)
              call merge_hkx3(tmp, EventFile_in(jj)%hkx3(kk), EventFile_out%hkx3(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)

      ! TODO add delete duplicates
  !        call EventFile_out%delete_duplicates_hk(status)

  end subroutine get_merged_hkx3_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the offset extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_offset_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_offset_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_offset(status, reopen)
      else
        call EventFile(ii)%read_offset(status)
      end if
    end do

  end subroutine read_all_offset_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge offset from infiles to outfiles
  !> also creates hdu of offset in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_offset_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tOffset) :: tmp

    character(*), parameter :: function = "get_merged_offset_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the offset extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_offset == false) then
        call alloc_offset(EventFile_out%offset, 0, status)
        EventFile_out%ext_offset = false
        return
      end if
    end do
    EventFile_out%ext_offset = true


    ! allocate the number of offset extensions in outfile
    call alloc_offset(EventFile_out%offset, EventFile_out%ninst, status)

    ! loop over all possible offset extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%offset(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        !                if(size(EventFile_in(jj)%offset) == 0) cycle
        ! loop over their corresponding offset extensions
        do kk = 1, size(EventFile_in(jj)%offset)
          ! check if the infile offset extension fits the out offset extension
          if(EventFile_in(jj)%offset(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%offset(ii) = EventFile_in(jj)%offset(kk)
            else
              tmp = EventFile_out%offset(ii)
              call merge_offset([tmp, EventFile_in(jj)%offset(kk)], EventFile_out%offset(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)
    call EventFile_out%delete_duplicates_offset(status)

  end subroutine get_merged_offset_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the badpix extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_badpix_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_badpix_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_badpix(status, reopen)
      else
        call EventFile(ii)%read_badpix(status)
      end if
    end do

  end subroutine read_all_badpix_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge badpix from infiles to outfiles
  !> if the EventFile type has no allocated badpix type, it
  !> also creates hdu of badpix in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_badpix_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    integer, intent(inout) :: status

    type(tbadpix) :: tmp
    integer :: ii, jj, kk, nfiles, first
    character(*), parameter :: function = "get_merged_badpix_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the badpix extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_badpix == false) then
        call alloc_badpix(EventFile_out%badpix, 0, status)
        EventFile_out%ext_badpix = false
        return
      end if
    end do
    EventFile_out%ext_badpix = true

    ! allocate the number of badpix extensions in outfile
    call alloc_badpix(EventFile_out%badpix, EventFile_out%ninst, status)

    ! loop over all possible badpix extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
        ! init all with zero rows
      call EventFile_out%badpix(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        if(size(EventFile_in(jj)%badpix) == 0) cycle
        ! loop over their corresponding badpix extensions
        do kk = 1, size(EventFile_in(jj)%badpix)
          ! check if the infile badpix extension fits the out badpix extension
          if(EventFile_in(jj)%badpix(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%badpix(ii) = EventFile_in(jj)%badpix(kk)
            !    call wrlog(function, "copy badpix header with ffits routines, but should be general routine")
            !                            call ffits_movnam_hdu(EventFile_in(jj)%unit, BINARY_TBL, EventFile_in(jj)%badpix(kk)%extname, status)
            !                            call ffits_copy_header(EventFile_in(jj)%unit, EventFile_out%unit, status)
            else
              tmp = EventFile_out%badpix(ii)
              call merge_badpix([tmp, EventFile_in(jj)%badpix(kk)], EventFile_out%badpix(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)
    call EventFile_out%delete_duplicates_badpix(status)

  end subroutine get_merged_badpix_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the deadcor extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_deadcor_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_deadcor_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_deadcor(status, reopen)
      else
        call EventFile(ii)%read_deadcor(status)
      end if
    end do

  end subroutine read_all_deadcor_eventfile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the events extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_events_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_events_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_events(status, reopen=reopen, use_wcs=true)
      else
        call EventFile(ii)%read_events(status, use_wcs=true)
      end if
    end do

  end subroutine read_all_events_eventfile

  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge deadcor from infiles to outfiles
  !> also creates hdu of deadcor in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_deadcor_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tDeadcor) :: tmp

    character(*), parameter :: function = "get_merged_deadcor_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)

    ! check that each input eventfile has the deadcor extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_deadcor == false) then
        call alloc_deadcor(EventFile_out%deadcor, 0, status)
        EventFile_out%ext_deadcor = false
        return
      end if
    end do
    EventFile_out%ext_deadcor = true


    ! allocate the number of deadcor extensions in outfile
    call alloc_deadcor(EventFile_out%deadcor, EventFile_out%ninst, status)

    ! loop over all possible deadcor extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
       ! init all with zero rows
      call EventFile_out%deadcor(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        if(size(EventFile_in(jj)%deadcor) == 0) cycle
        ! loop over their corresponding deadcor extensions
        do kk = 1, size(EventFile_in(jj)%deadcor)
          ! check if the infile deadcor extension fits the out deadcor extension
          if(EventFile_in(jj)%deadcor(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%deadcor(ii) = EventFile_in(jj)%deadcor(kk)
            else
              tmp = EventFile_out%deadcor(ii)
              call merge_deadcor([tmp, EventFile_in(jj)%deadcor(kk)], EventFile_out%deadcor(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)
    call EventFile_out%delete_duplicates_deadcor(status)

  end subroutine get_merged_deadcor_eventfile
  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> read the corrat extensions of all infiles
  !
  !
  !
  !> @param[in, optional] reopen
  !> @param[out]
  !> @param[inout] status, EventFile
  !---------------------------------------------------------------------------

  subroutine read_all_corrat_eventfile(EventFile, status, reopen)
    implicit none

    integer, intent(inout) :: status
    integer :: ii
    integer, intent(in), optional :: reopen

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "read_all_corrat_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      if(present(reopen)) then
        call EventFile(ii)%read_corrat(status, reopen)
      else
        call EventFile(ii)%read_corrat(status)
      end if
    end do

  end subroutine read_all_corrat_eventfile


  !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge all corrat from infiles to outfiles
  !> also creates hdu of corrat in outfile
  !
  !
  !> @param[in] EventFile_in
  !> @param[out] EventFile_out
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine get_merged_corrat_eventfile(EventFile_in, EventFile_out, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii, jj, kk, nfiles, first

    type(tEventFile), dimension(:), intent(inout) :: EventFile_in
    type(tEventFile), intent(inout) :: EventFile_out
    type(tCorrat) :: tmp
    logical, dimension(:), allocatable :: extensionIsAbsent
    character(buffsize) :: msg
    character(*), parameter :: function = "get_merged_corrat_eventfile"

    if(status > 100 .or. status == s_errRep) return

    nfiles = size(EventFile_in)
    call alloc(extensionIsAbsent, nfiles, status)
    extensionIsAbsent = .false. ! default

    ! check that each input eventfile has the corrat extensions - to create the merged extensions
    do ii = 1, nfiles
      if(EventFile_in(ii)%ext_corrat == false) then
        extensionIsAbsent(ii) = .true.
      end if
    end do

    if(any(extensionIsAbsent)) then
      if(all(extensionIsAbsent)) then
        msg = "No input file has any CORRATT extensions."
      else
        msg = "At least one of the input files doesn't have CORRATT extensions, thus we can't create one in the output."
      end if
      call wrerr(function, warn, msg)

      call alloc_corrat(EventFile_out%corrat, 0, status)
      EventFile_out%ext_corrat = false
      call dealloc(extensionIsAbsent, status)
  return
    end if

    EventFile_out%ext_corrat = true
    call dealloc(extensionIsAbsent, status)

    ! allocate the number of corrat extensions in outfile
    call alloc_corrat(EventFile_out%corrat, EventFile_out%ninst, status)

    ! loop over all possible corrat extension in the outfile
    do ii = 1, EventFile_out%ninst
      first = true
      ! init all with zero rows
      call EventFile_out%corrat(ii)%init(0, status)
      ! loop over all infiles
      do jj = 1, nfiles
        if(size(EventFile_in(jj)%corrat) == 0) cycle
        ! loop over their corresponding corrat extensions
        do kk = 1, size(EventFile_in(jj)%corrat)
          ! check if the infile corrat extension fits the out corrat extension
          if(EventFile_in(jj)%corrat(kk)%ccdnr == EventFile_out%ccdnr_infile(ii)) then
            if(first == true) then
              first = false
              EventFile_out%corrat(ii) = EventFile_in(jj)%corrat(kk)
            else
              tmp = EventFile_out%corrat(ii)
              call merge_Corrat([tmp, EventFile_in(jj)%corrat(kk)], EventFile_out%corrat(ii), status)
            end if
          end if
        end do
      end do
    end do

    call tmp%free(status)
    call EventFile_out%delete_duplicates_corrat(status)

  end subroutine get_merged_corrat_eventfile

    !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the corrat extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_corrat_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_corrat_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_corrat(status)
    end do

  end subroutine free_all_corrat_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the deadcor extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_deadcor_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_deadcor_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_deadcor(status)
    end do

  end subroutine free_all_deadcor_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the events extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_events_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_events_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_events(status)
    end do

  end subroutine free_all_events_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the badpix extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_badpix_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_badpix_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_badpix(status)
    end do

  end subroutine free_all_badpix_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hk extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_hk_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_hk_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_hk(status)
    end do

  end subroutine free_all_hk_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx1 extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_hkx1_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_hkx1_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_hkx1(status)
    end do

  end subroutine free_all_hkx1_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx2 extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_hkx2_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_hkx2_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_hkx2(status)
    end do

  end subroutine free_all_hkx2_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx3 extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_hkx3_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_hkx3_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_hkx3(status)
    end do

  end subroutine free_all_hkx3_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the hkx5 extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_hkx5_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_hkx5_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_hkx5(status)
    end do

  end subroutine free_all_hkx5_eventfile

          !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> free the offset extensions of all infiles
  !
  !
  !
  !> @param[in]
  !> @param[out] outParam
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine free_all_offset_eventfile(EventFile, status)
    implicit none

    integer, intent(inout) :: status
    integer :: ii

    type(tEventFile), dimension(:), intent(inout) :: EventFile

    character(*), parameter :: function = "free_all_offset_eventfile"

    if(status > 100 .or. status == s_errRep) return

    do ii = 1, size(EventFile)
      call EventFile(ii)%free_offset(status)
    end do

  end subroutine free_all_offset_eventfile

      !---------------------------------------------------------------------------
  !> @author
  !> Christoph Grossberger
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merge arrays of telescope ids and their corresponding instruments
  !
  !
  !
  !> @param[in] array1, array2, array1_str, array2_str
  !> @param[out] outarray, outarray_str
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_ccdnr(array1, array2, outarray, array1_str, array2_str, outarray_str, status)
    implicit none

    integer, intent(inout) :: status
    integer(i4), dimension(:), intent(in) :: array1, array2
    integer(i4), allocatable, dimension(:), intent(out) :: outarray
!**** 2019-09-11 IMS: all the above variables are declared integer in the calling routine, not i4.
    integer(i4), dimension(size(array1) + size(array2)) :: tmp
    integer, allocatable, dimension(:) :: order

    character(buffsize), dimension(:), intent(in) :: array1_str
    character(buffsize), dimension(:), intent(in) :: array2_str
    character(buffsize), allocatable, dimension(:), intent(out) :: outarray_str
    character(buffsize), dimension(size(array1) + size(array2)) :: tmp_str

    logical, allocatable, dimension(:) :: mask
    integer :: ii

    character(*), parameter :: function = "merge_ccdnr"

    if(status > 100 .or. status == s_errRep) return

    call alloc(mask, size(array1) + size(array2), status)

    tmp = [array1, array2]
    tmp_str = [array1_str, array2_str]

    call sort(tmp, status, order)
    tmp_str = tmp_str(order)

    mask = .true.
    do ii = 2, size(tmp)
      if(tmp(ii) == tmp(ii-1)) mask(ii) = .false.
    end do

    call alloc(outarray, size(array1) + size(array2), status)
    call alloc(outarray_str, size(array1) + size(array2), status)

    outarray = pack(tmp, mask)
    outarray_str = pack(tmp_str, mask)

    call dealloc(order, status)
    call dealloc(mask, status)

  end subroutine merge_ccdnr

      !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> compares RAWXY selection images.
  !
  !
  !
  !> @param[in] array1, array2, array1_str, array2_str
  !> @param[out] outarray, outarray_str
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine compare_rawxy_masks(ccdnrs1, rawxy_mask1, ccdnrs2, rawxy_mask2, status)
    implicit none

    integer, dimension(:), intent(in) :: ccdnrs1, ccdnrs2
    type(tRawXYMask), intent(in) :: rawxy_mask1, rawxy_mask2
    integer, intent(inout) :: status

    character(*), parameter :: function = "compare_rawxy_masks"

    logical :: is_alloc1,is_alloc2
    integer :: ii,jj,ccdnr
    logical, dimension(n_ccd_pixels,n_ccd_pixels) :: mask1,mask2

    if(status > 100 .or. status == s_errRep) return

    ! What we don't want to allow is if evlist 1 and evlist 2 share a CCD number,
    ! but have different rawxy masks for that CCD. (Note that if one mask is not
    ! allocated, and the other mask is all .true. for the relevant CCD, we will
    ! allow that to pass.
    !
    is_alloc1 = allocated(rawxy_mask1%values)
    is_alloc2 = allocated(rawxy_mask2%values)

    if(.not. (is_alloc1 .or. is_alloc2)) return

!*** I am assuming here that there are no repeated elements in either ccdnrs1 or ccdnrs2.
    do ii=1,size(ccdnrs1)
      ccdnr = ccdnrs1(ii)
      do jj=1,size(ccdnrs2)
        if(ccdnr==ccdnrs2(jj)) then
          if(is_alloc1) then
            mask1 = btest(rawxy_mask1%values, ccdnr-1)
            if(is_alloc2) then
              mask2 = btest(rawxy_mask2%values, ccdnr-1)
              if(.not. all(mask1 .eqv. mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            else
              if(.not. all(mask1)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            end if

          else ! then rawxy_mask2 is allocated
            mask2 = btest(rawxy_mask2%values, ccdnr-1)
            if(is_alloc1) then
              mask1 = btest(rawxy_mask1%values, ccdnr-1)
              if(.not. all(mask1 .eqv. mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            else
              if(.not. all(mask2)) then
                call wrerr(function, s_err, "RawXY selection masks not equivalent for 2 input event lists.")
                status = s_errRep
  return
              end if
            end if
          end if
        end if
      end do
    end do

  end subroutine compare_rawxy_masks

      !---------------------------------------------------------------------------
  !> @author
  !> Ian M Stewart
  !
  ! DESCRIPTION:
  !> Brief description of routine.
  !> @brief
  !> merges RAWXY selection images.
  !
  !
  !
  !> @param[in] array1, array2, array1_str, array2_str
  !> @param[out] outarray, outarray_str
  !> @param[inout] status
  !---------------------------------------------------------------------------

  subroutine merge_rawxy_masks(ccdnrs2, rawxy_mask2, rawxy_mask, ccd_is_merged, status)
    implicit none

    integer, dimension(:), intent(in)    :: ccdnrs2
    type(tRawXYMask),       intent(in)    :: rawxy_mask2
    type(tRawXYMask),       intent(inout) :: rawxy_mask
    logical, dimension(:), intent(inout) :: ccd_is_merged
    integer,               intent(inout) :: status

    character(*), parameter :: function = "merge_rawxy_masks"

    integer :: ii,ccdnr,xi,yi
    logical, dimension(n_ccd_pixels,n_ccd_pixels) :: mask

    if(status > 100 .or. status == s_errRep) return

    do ii=1,size(ccdnrs2)
      ccdnr = ccdnrs2(ii)
      if(.not.ccd_is_merged(ccdnr)) then
        if(allocated(rawxy_mask2%values)) then
          if(.not.allocated(rawxy_mask%values)) then
            call rawxy_mask%init(status)
            rawxy_mask%extname = rawxy_mask2%extname
          end if

          mask = btest(rawxy_mask2%values, ccdnr-1)
! 2019-09-11 IMS: note that overload to int8 for all bit manipulation functions appears to be gnu-specific (i.e., non-standard f90). Will it work on all O/S?

          ! IMS: the following will surely be slow, but so far I can't think of a funky array method.
          !
          do xi=1,n_ccd_pixels
            do yi=1,n_ccd_pixels
              if(mask(xi,yi)) then
                rawxy_mask%values(xi,yi) = ibset(rawxy_mask%values(xi,yi), ccdnr-1)
              else
                rawxy_mask%values(xi,yi) = ibclr(rawxy_mask%values(xi,yi), ccdnr-1)
              end if
            end do
          end do
        end if

        ccd_is_merged(ccdnr) = .true.
      end if
    end do

  end subroutine merge_rawxy_masks

end module evtool_eventfile_routines_mod
